from orderly_evm_connector.lib.utils import check_required_parameters


def create_referral_code(self, account_id: str, referral_code: str, max_rebate_rate: float,
                         referrer_rebate_rate: float, referee_rebate_rate: float):
    """
    Create Referral Code

    Limit: 1 requests per 1 second

    POST /v1/referral/create

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/create-referral-code
    """
    check_required_parameters(
        [[account_id, "account_id"],
         [referral_code, "referral_code"],
         [max_rebate_rate, "max_rebate_rate"],
         [referrer_rebate_rate, "referrer_rebate_rate"],
         [referee_rebate_rate, "referee_rebate_rate"]]
    )
    payload = {
        "account_id": account_id,
        "referral_code": referral_code,
        "max_rebate_rate": max_rebate_rate,
        "referrer_rebate_rate": referrer_rebate_rate,
        "referee_rebate_rate": referee_rebate_rate
    }
    return self._sign_request("POST", "/v1/referral/create", payload=payload)


def update_referral_code(self, account_id: str, referral_code: str, max_rebate_rate: float,
                         referrer_rebate_rate: float, referee_rebate_rate: float):
    """
    Update Referral Code

    Limit: 1 requests per 1 second

    POST /v1/referral/ypdate

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/update-referral-code
    """
    check_required_parameters(
        [[account_id, "account_id"],
         [referral_code, "referral_code"],
         [max_rebate_rate, "max_rebate_rate"],
         [referrer_rebate_rate, "referrer_rebate_rate"],
         [referee_rebate_rate, "referee_rebate_rate"]]
    )
    payload = {
        "account_id": account_id,
        "referral_code": referral_code,
        "max_rebate_rate": max_rebate_rate,
        "referrer_rebate_rate": referrer_rebate_rate,
        "referee_rebate_rate": referee_rebate_rate
    }
    return self._sign_request("POST", "/v1/referral/update", payload=payload)


def bind_referral_code(self, referral_code: str):
    """
    Bind Referral Code

    Limit: 1 requests per 1 second

    POST /v1/referral/bind

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/bind-referral-code
    """
    check_required_parameters(
        [[referral_code, "referral_code"]]
    )
    payload = {
        "referral_code": referral_code
    }
    return self._sign_request("POST", "/v1/referral/bind", payload=payload)


def get_referral_code_info(self, page: int = None, size: int = None, user_address: str = None, account_id : str = None):
    """
    [Private]Get Referral Code Info
    
    Scope: Only each broker_id’s admin wallet can call this endpoint.


    Limit: 10 requests per 1 second

    GET /v1/referral/admin_info

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-referral-code-info
    """
    payload = {
        "page": page,
        "size": size,
    }
    if user_address:
        payload["user_address"] = user_address
    if account_id:
        payload["account_id"] = account_id
    return self._sign_request("GET", "/v1/referral/admin_info", payload=payload)


def get_referral_info(self):
    """
    Get Referral Info

    Limit: 10 requests per 1 second

    GET /v1/referral/info

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/get-referral-info
    """
    return self._sign_request("GET", "/v1/referral/info")


def get_referral_history(self, start_date: str = None, end_date: str = None, page: int = None, size: int = None):
    """
    Get Referral History

    Limit: 10 requests per 1 second

    GET /v1/referral/history

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/get-referral-history
    """
    payload = {
        "start_date": start_date,
        "end_date": end_date,
        "page": page,
        "size": size,
    }
    return self._sign_request("GET", "/v1/referral/referral_history", payload=payload)


def get_referral_rebate_summary(self, start_date: str = None, end_date: str = None, page: int = None, size: int = None):
    """
    Get Referral Rebate Summary

    Limit: 10 requests per 1 second

    GET /v1/referral/rebate_summary
    
    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-referral-rebate-summary#openapi-evmopenapi-get-v1referralrebate_summary
    """
    print(start_date,end_date)
    check_required_parameters([[start_date,"start_date"],[end_date,"end_date"]])
    payload = {
        "start_date": start_date,
        "end_date": end_date,
        "page": page,
        "size": size,
    }
    return self._sign_request("GET", "/v1/referral/rebate_summary", payload=payload)


def get_referee_history(self, start_date: str = None, end_date: str = None, page: int = None, size: int = None):
    """
    Get Referee History

    Limit: 10 requests per 1 second

    GET /v1/referral/referee_history

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/referee_history
    """
    payload = {
        "start_date": start_date,
        "end_date": end_date,
        "page": page,
        "size": size,
    }
    return self._sign_request("GET", "/v1/referral/referee_history", payload=payload)


def get_referee_info(self, sort: str = None, page: int = None, size: int = None):
    """
    Get Referee Info

    Limit: 10 requests per 1 second

    GET /v1/referral/referee_info

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/get-referee-info
    """
    payload = {
        "sort": sort,
        "page": page,
        "size": size,
    }
    return self._sign_request("GET", "/v1/referral/referee_info", payload=payload)


def get_distribution_history(self, start_t: str = None, end_t: str = None, page: int = None, size: int = None,type: str = None, status: str = None):
    """
    Get Distribution History

    Limit: 1 requests per 1 second

    GET /v1/client/distribution_history

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/get-distribution-history
    """
    payload = {
        "start_t": start_t,
        "end_t": end_t,
        "page": page,
        "size": size,
        "type": type,
        "status": status
    }
    return self._sign_request("GET", "/v1/client/distribution_history", payload=payload)


def check_ref_code(self, account_id:str = None ):
    """
    Check Referral Code

    Limit: 10 requests per second 
    GET /v1/public/referral/check_ref_code

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/check-referral-code#openapi-evmopenapi-get-v1publicreferralcheck_ref_code
    """
    check_required_parameters([[account_id,'account_id']])
    return self._request("GET", f"/v1/public/referral/check_ref_code?account_id={account_id}")

def verify_ref_code(self, referral_code:str = None ):
    """
    Verify Referral Code

    Limit: 10 requests per second 
    GET /v1/public/referral/verify_ref_code
    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/verify-referral-code#openapi-evmopenapi-get-v1publicreferralverify_ref_code
    """
    check_required_parameters([[referral_code,'referral_code']])
    return self._request("GET", f"/v1/public/referral/verify_ref_code?referral_code={referral_code}")

def edit_referral_split(self, referral_code: str, referrer_rebate_rate: float, referee_rebate_rate: float):
    """
    Edit Split

    Limit: 10 requests per 1 second

    POST /v1/referral/edit_split

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/private/edit-split
    """
    check_required_parameters([[referral_code,'referral_code'], [referrer_rebate_rate,'referrer_rebate_rate'], [referee_rebate_rate,'referee_rebate_rate']])
    return self._sign_request("POST", "/v1/referral/edit_split", payload={
        "referral_code": referral_code,
        "referrer_rebate_rate": referrer_rebate_rate,
        "referee_rebate_rate": referee_rebate_rate
    })


import json
from json import JSONDecodeError
import requests
from .__version__ import __version__
from orderly_evm_connector.error import ClientError, ServerError
from orderly_evm_connector.lib.utils import (
    generate_signature,
    generate_wallet_signature,
)
from orderly_evm_connector.lib.utils import cleanNoneValue
from orderly_evm_connector.lib.utils import orderlyLog, get_endpoints

class API(object):
    def __init__(
        self,
        orderly_key=None,
        orderly_secret=None,
        wallet_secret=None,
        orderly_testnet=False,
        orderly_account_id=None,
        proxies=None,
        timeout=None,
        debug=False
    ):
        self.orderly_key = orderly_key
        self.orderly_secret = orderly_secret
        self.wallet_secret = wallet_secret
        self.orderly_endpoint, _, _ = get_endpoints(orderly_testnet)
        self.orderly_account_id = orderly_account_id
        self.timeout = timeout
        self.show_header = False
        self.proxies = proxies
        self.logger = orderlyLog(debug=debug)
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Content-Type": "application/json;charset=utf-8",
                "User-Agent": "orderly-connector-python/" + __version__,
            }
        )
        return

    def _request(self, http_method, url_path, payload=None):
        if payload:
            _payload = cleanNoneValue(payload)
            if _payload:
                if http_method == "GET" or http_method == "DELETE":
                    url_path += "?" + "&".join(
                        [f"{k}={v}" for k, v in _payload.items()]
                    )
                    payload = ""
                else:
                    payload = _payload

        if payload is None:
            payload = ""
        url = self.orderly_endpoint + url_path
        self.logger.debug("url: " + url)
        params = cleanNoneValue(
            {
                "url": url,
                "params": payload,
                "timeout": self.timeout,
                "proxies": self.proxies,
            }
        )
        response = self._dispatch_request(http_method, params)
        self.logger.debug("raw response from server:" + response.text)
        self._handle_rest_exception(response)

        try:
            data = response.json()
        except ValueError:
            data = response.text

        return data

    def get_wallet_signature(self, message=None):
        return generate_wallet_signature(self.wallet_secret, message=message)

    def _sign_request(self, http_method, url_path, payload=None):
        _payload = ""
        if payload:
            _payload = cleanNoneValue(payload)
            if _payload:
                if http_method == "GET" or http_method == "DELETE":
                    url_path += "?" + "&".join(
                        [f"{k}={v}" for k, v in _payload.items()]
                    )
                    _payload = ""
        params = {}
        payload = _payload if _payload else ""
        params["url_path"] = url_path
        params["payload"] = payload
        params["http_method"] = http_method
        query_string = self._prepare_params(params)
        try:
            _timestamp, _signature = generate_signature(
                self.orderly_secret, message=query_string
            )
        except ValueError:
            _timestamp, _signature = "mock_timestamp", "mock_signature"

        self.session.headers.update(
            {
                "orderly-timestamp": _timestamp,
                "orderly-account-id": self.orderly_account_id,
                "orderly-key": self.orderly_key,
                "orderly-signature": _signature,
            }
        )
        self.logger.debug(f"Sign Request Headers: {self.session.headers}")
        return self.send_request(http_method, url_path, payload)

    def send_request(self, http_method, url_path, payload=None):
        if payload is None:
            payload = {}
        url = self.orderly_endpoint + url_path
        self.logger.debug("url: " + url)
        params = cleanNoneValue(
            {
                "url": url,
                "params": payload,
                "timeout": self.timeout,
                "proxies": self.proxies,
            }
        )
        response = self._dispatch_request(http_method, params)
        self.logger.debug("raw response from server:" + response.text)
        self._handle_rest_exception(response)

        try:
            data = response.json()
        except ValueError:
            data = response.text
        result = {}

        if self.show_header:
            result["header"] = response.headers

        if len(result) != 0:
            result["data"] = data
            return result
        return data

    def _prepare_params(self, params: dict):
        _http_method = params["http_method"]
        _url_path = params["url_path"]
        _payload = (
            json.dumps(params["payload"]) if params["payload"] else params["payload"]
        )
        _params = "{0}{1}{2}".format(_http_method, _url_path, _payload)
        return _params

    def _dispatch_request(self, http_method, params):
        method_func = {
            "GET": self.session.get,
            "DELETE": self.session.delete,
            "PUT": self.session.put,
            "POST": self.session.post,
        }.get(http_method, "GET")
        if http_method == "POST" or http_method == "PUT":
            self.session.headers.update(
                {
                    "Content-Type": "application/json;charset=utf-8"
                }
            )
            return method_func(url=params["url"], json=params["params"])
        else:
            self.session.headers.update(
                {
                    "Content-Type": "application/x-www-form-urlencoded;charset=utf-8"
                }
            )
            return method_func(url=params["url"])

    def _handle_rest_exception(self, response):
        status_code = response.status_code
        if status_code <= 400:
            return
        if 400 < status_code < 500:
            try:
                err = json.loads(response.text)
            except JSONDecodeError:
                raise ClientError(
                    status_code, None, response.text, None, response.headers
                )
            error_data = None
            if "data" in err:
                error_data = err["data"]
            raise ClientError(
                status_code, err["code"], err["message"], response.headers, error_data
            )
        raise ServerError(status_code, response.text)



from orderly_evm_connector.lib.utils import check_required_parameters


def get_all_notifications(self, **kwargs):
    """[Private] Get all notifications

    Limit: 10 requests per 60 seconds

    Get all notifications for the user.

    Currently the only supported message type is the order filled message.

    GET /v1/notification/inbox/notifications

    Optional Args:
        type(string): Filter nofications by type (TRADE / SYSTEM).
        page(number): (default: 1)  the page you wish to query.
        size(number): (default: 25) the page size you wish to query. (max: 500)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-all-notifications

    """
    payload = {**kwargs}
    return self._sign_request(
        "GET", "/v1/notification/inbox/notifications", payload=payload
    )


def get_unread_notifications(self):
    """[Private] Get unread notification information

    Limit: 10 requests per 60 seconds

    GET /v1/notification/inbox/unread

    Get the information on unread messages for the user.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-unread-notifications
    """
    return self._sign_request("GET", "/v1/notification/inbox/unread")


def set_read_status_notifications(self, flag: int, ids: list):
    """[Private] Set read status of notifications

    Limit: 10 requests per 60 seconds

    POST /v1/notification/inbox/mark_read

    Set the read status on a list of notifications for a user.

    Args:
        flag(number): The value of the read flag, where 1 = READ and 0 = UNREAD
        ids(list): The list of notification ids to flag as read/unread.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/set-read-status-of-notifications
    """
    check_required_parameters([[flag, "flag"], [ids, "ids"]])
    payload = {"flag": flag, "ids": ids}
    return self._sign_request(
        "POST", "/v1/notification/inbox/mark_read", payload=payload
    )


def set_read_status_all_notifications(self, flag: int):
    """[Private] Set read status of all notifications

    Limit: 10 requests per 60 seconds

    POST /v1/notification/inbox/mark_read_all

    Set the read status on all notifications for a user.

    Args:
        flag(number): The value of the read flag, where 1 = READ and 0 = UNREAD

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/set-read-status-of-all-notifications
    """
    payload = {"flag": flag}
    check_required_parameters([[flag, "flag"]])
    return self._sign_request(
        "POST", "/v1/notification/inbox/mark_read_all", payload=payload
    )



from orderly_evm_connector.lib.utils import check_required_parameters


def get_system_maintenance_status(self):
    """[Public] System maintenance status

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/system_info

    Retreive the current system maintenance status of Orderly Network. A return value of status = 0 means the system is functioning properly and a return value of status = 2 means the system is under maintenance.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-system-maintenance-status
    """
    return self._request("GET", "/v1/public/system_info")


def get_faucet_usdc(self, chain_id: str, user_address: str):
    """[Public] Get faucet USDC(Testnet only)

    Receive 1,000 USDC in the Testnet environment. Each account may only use the faucet a maximum of 5 times.

    POST https://testnet-operator-evm.orderly.org/v1/faucet/usdc

    Args:
    chain_id(string): The chain ID that the test USDC should be deposited to
    user_address(string): The address of the user account

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-faucet-usdctestnet-only
    """

    check_required_parameters([[chain_id, "chain_id"], [user_address, "user_address"]])

    self.orderly_endpoint = "https://testnet-operator-evm.orderly.org"
    payload = {
        "broker_id": "woo-dex",
        "chain_id": chain_id,
        "user_address": user_address,
    }
    self.logger.info(
        f"Receive 1,000 USDC in the Testnet environment. {self.orderly_endpoint} {payload}"
    )
    return self._request("POST", "/v1/faucet/usdc", payload=payload)


def get_exchange_info(self, symbol: str):
    """[Public] Exchange information

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/info/:symbol

    This endpoint provides all the values for the rules that an order need to fulfil in order for it to be placed successfully. The rules are defined as follows:

    Price filter

    price >= quote_min
    price <= quote_max
    (price - quote_min) % quote_tick should equal to zero
    price <= asks[0].price * (1 + price_range) when BUY
    price >= bids[0].price * (1 - price_range) when SELL
    Size filter

    base_min <= quantity <= base_max
    (quantity - base_min) % base_tick should equal to zero
    Min Notional filter

    price * quantity should greater than min_notional
    Risk Exposure filer

    Order size should be within holding threshold. See account_info

    Get available symbols that Orderly Network supports, and also send order rules for each symbol. The definition of rules can be found at Exchange Infomation

    Args:
        symbol(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-exchange-information
    """
    check_required_parameters([[symbol, "symbol"]])
    return self._request("GET", f"/v1/public/info/{symbol}")


def get_token_info(self, chain_id: str = None):
    """[Public] Token info

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/token

    Retrives the available tokens to custody within Orderly Network.

    Args:
        chain_id: str

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-supported-collateral-info

    """

    payload = {
        "chain_id": chain_id
    }
    return self._request("GET", "/v1/public/token", payload=payload)


def get_available_symbols(self):
    """[Public] Available symbols

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/info

    Get available symbols that Orderly Network supports, and also send order rules for each symbol. The definition of rules can be found at Exchange Infomation

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-available-symbols
    """
    return self._request("GET", "/v1/public/info")


def get_fee_futures_information(self):
    """[Public] Futures fee information

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/fee_futures/program

    Get fee information for futures trading.

    https://docs-api-evm.orderly.network/#restful-api-public-futures-fee-information
    """
    return self._request("GET", "/v1/public/fee_futures/program")


def get_leverage_configuration(self):
    """[Public] Get leverage configuration

    Limit: 10 requests per 1 second per IP address

    GET v1/public/config

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-leverage-configuration
    """
    return self._request("GET", "/v1/public/config")


def get_user_statistics(self):
    """[Public] Get user statistics
    Limit: 10 requests per 60 seconds

    GET /v1/client/statistics

    Get statistics of the user account

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-user-statistics

    """
    return self._sign_request("GET", "/v1/client/statistics")


def get_market_volume_by_broker(self, symbol: str = None, broker_id: str = None):
    """[Public] Get market volume by broker
    Limit: 10 requests per 60 seconds

    GET /v1/public/futures_market

    Get market volume filtered by broker

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-market-volume-by-broker

    """

    payload = {
        "symbol": symbol,
        "broker_id": broker_id
    }
    return self._request("GET", "/v1/public/futures_market", payload=payload)


import asyncio
import inspect
import json
import websockets
from websockets.exceptions import (
    ConnectionClosedError,
    ConnectionClosedOK,
    WebSocketException,
)
from orderly_evm_connector.lib.utils import orderlyLog, parse_proxies, decode_ws_error_code
from orderly_evm_connector.lib.constants import (
    WEBSOCKET_TIMEOUT_IN_SECONDS,
    WEBSOCKET_FAILED_MAX_RETRIES,
    WEBSOCKET_RETRY_SLEEP_TIME,
)

class AsyncWebsocketManager:
    def __init__(
        self,
        websocket_url,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
        on_ping=None,
        on_pong=None,
        timeout=WEBSOCKET_TIMEOUT_IN_SECONDS,
        debug=False,
        proxies=None,
        max_retries=WEBSOCKET_FAILED_MAX_RETRIES,
    ):
        self.websocket_url = websocket_url
        self.on_message = on_message
        self.on_open = on_open
        self.on_close = on_close
        self.init = False
        self.on_error = on_error
        self.on_ping = on_ping
        self.on_pong = on_pong
        self.timeout = timeout
        self.logger = orderlyLog(debug=debug)
        self._proxy_params = parse_proxies(proxies) if proxies else {}
        self.subscriptions = []
        self._login = False
        self.max_retries = max_retries
        self.ws = None
        self.loop = asyncio.get_event_loop()
        self._stopping = False
        self._on_close_called = False

    def start(self):
        pass

    async def create_ws_connection(self):
        retries = 0
        while retries <= self.max_retries:
            try:
                self.logger.debug(
                    f"Creating connection with WebSocket Server: {self.websocket_url}, proxies: {self._proxy_params}"
                )
                self.ws = await websockets.connect(
                    self.websocket_url, timeout=self.timeout, **self._proxy_params
                )
                self.logger.debug(
                    f"WebSocket connection has been established: {self.websocket_url}, proxies: {self._proxy_params}"
                )
                self._stopping = False
                self._on_close_called = False
                self.init = False
                if self.on_open:
                    await self._callback(self.on_open)
                return
            except Exception as e:
                self.logger.error(f"Failed to create WebSocket connection: {e}")
                retries += 1
                if retries <= self.max_retries:
                    self.logger.warning(
                        f"Retrying connection... (Attempt {retries}/{self.max_retries})"
                    )
                    await asyncio.sleep(WEBSOCKET_RETRY_SLEEP_TIME)
                else:
                    raise

    async def reconnect(self):
        if self._stopping:
            return
        self.logger.warning("Reconnecting to WebSocket...")
        await self._internal_close()
        self._login = False
        await self.create_ws_connection()

    def send_message(self, message):
        self.logger.debug("Sending message to Orderly WebSocket Server: %s", message)
        if self.ws and not self.ws.closed and not self._stopping:
            asyncio.create_task(self.ws.send(message))
        else:
            self.logger.debug("Skipping send_message because websocket is not connected.")

    async def run(self):
        await self.create_ws_connection()
        await self.read_data()

    async def ensure_init(self):
        while not self._stopping:
            if self.init:
                break
            await asyncio.sleep(1)

    async def _handle_heartbeat(self):
        try:
            _payload = {"event": "pong"}
            if self.ws and not self.ws.closed:
                await self.ws.send(json.dumps(_payload))
                self.logger.debug(f"Sent Ping frame: {_payload}")
        except Exception as e:
            self.logger.error("Failed to send Ping: {}".format(e))

    async def read_data(self):
        while True:
            if self._stopping:
                await self._notify_close()
                break
            try:
                ws = self.ws
                if ws is None:
                    await asyncio.sleep(WEBSOCKET_RETRY_SLEEP_TIME)
                    continue
                message = await ws.recv()
                self.init = True
            except (ConnectionClosedError, ConnectionClosedOK) as e:
                close_code = getattr(e, "code", None)
                close_reason = getattr(e, "reason", "")
                if self._stopping:
                    self.logger.info("WebSocket connection closed by client.")
                    await self._notify_close()
                    break
                if close_code == 1000:
                    self.logger.info("WebSocket connection closed normally (code 1000).")
                else:
                    self.logger.warning(
                        f"WebSocket connection closed (code={close_code}, reason={close_reason})."
                    )
                await self._notify_close()
                await self.reconnect()
                continue
            except WebSocketException as e:
                if self._stopping:
                    break
                self.logger.error(f"WebSocket exception: {e}")
                await self._notify_close()
                await self.reconnect()
                continue
            except Exception as e:
                if self._stopping:
                    break
                self.logger.error(f"Exception in read_data: {e}")
                await self._notify_close()
                await self.reconnect()
                continue

            payload = b""
            try:
                payload = message
                if isinstance(message, str):
                    payload = message.encode()
                elif isinstance(message, (bytearray, memoryview)):
                    payload = bytes(message)
                _message = json.loads(message)
            except json.JSONDecodeError:
                err_code = decode_ws_error_code(payload)
                if err_code == "1000":
                    self.logger.info("Websocket closed normally (code 1000).")
                    if self._stopping:
                        await self._notify_close()
                        break
                elif err_code:
                    self.logger.warning(
                        f"Websocket error code received: {err_code}"
                    )
                continue

            if "event" in _message and _message["event"] == "ping":
                await self._handle_heartbeat()
            else:
                await self._callback(self.on_message, _message)

    async def close(self):
        self._stopping = True
        await self._internal_close()

    async def _callback(self, callback, *args):
        if callback:
            try:
                result = callback(self, *args)
                if inspect.isawaitable(result):
                    await result
            except Exception as e:
                self.logger.error("Error from callback {}: {}".format(callback, e))
                if self.on_error and callback is not self.on_error:
                    await self._callback(self.on_error, e)

    async def _notify_close(self):
        if not self._on_close_called:
            self._on_close_called = True
            await self._callback(self.on_close)

    async def _internal_close(self):
        if self.ws and not self.ws.closed:
            await self.ws.close()
        self.ws = None
        self.init = False
        self._login = False



from orderly_evm_connector.lib.utils import check_required_parameters


def get_parameters_of_each_epoch(self):
    """
    Get Parameters of Each Epoch of All Epochs

    Limit: 10 requests per 1 second

    GET /v1/public/trading_rewards/epoch_info

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-parameters-of-each-epoch-for-all-epochs
    """

    return self._request("GET", "/v1/public/trading_rewards/epoch_info")


def get_epochs_data(self):
    """
    Get Epochs Data

    Limit: 10 requests per 1 second

    GET /v1/public/trading_rewards/epoch_data

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-epochs-data
    """

    return self._request("GET", "/v1/public/trading_rewards/epoch_data")

def get_broker_allocation_history(self):
    """
    Get Broker Allocation History

    Limit: 10 requests per 1 second

    GET /v1/public/trading_rewards/broker_allocation_history

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-broker-allocation-history
    """

    return self._request("GET", "/v1/public/trading_rewards/broker_allocation_history")


def get_wallet_trading_rewards_history(self, address: str):
    """
    Get Wallet Trading Rewards History

    Limit: 10 requests per 1 second

    GET /v1/public/trading_rewards/wallet_rewards_history

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-wallet-trading-rewards-history
    """
    check_required_parameters([
        [address, "address"],
    ])
    payload = {
        "address": address,
    }
    return self._request("GET", "/v1/public/trading_rewards/wallet_rewards_history", payload=payload)


def get_account_trading_rewards_history(self, address: str):
    """
    Get Account Trading Rewards History

    Limit: 10 requests per 1 second

    GET /v1/public/trading_rewards/account_rewards_history

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-account-trading-rewards-history
    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address}
    return self._request("GET", "/v1/public/trading_rewards/account_rewards_history", payload=payload)


def get_current_epoch_estimate(self, address: str):
    """
    Get Current Epoch Estimate

    Limit: 10 requests per 1 second

    GET /v1/public/trading_rewards/current_epoch_estimate

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-current-epoch-estimate
    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address}
    return self._request("GET", "/v1/public/trading_rewards/current_epoch_estimate", payload=payload)


def get_current_epoch_estimate_broker(self):
    """
    Get Current Epoch Estimate by Broker

    Limit: 10 requests per 1 second

    GET /v1/public/trading_rewards/current_epoch_broker_estimate

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-current-epoch-estimate-by-broker
    """
    return self._request("GET", "/v1/public/trading_rewards/current_epoch_broker_estimate")



def get_parameters_of_each_mm_epoch(self):
    """
    Get Parameters of Each MM Epoch of All MM Epochs

    Limit: 10 requests per 1 second

    GET /v1/public/market_making_rewards/epoch_info

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-parameters-of-each-mm-epoch-for-all-mm-epochs
    """

    return self._request("GET", "/v1/public/market_making_rewards/epoch_info")


def get_wallet_group_mm_rewards_history(self, address: str, symbol: str = None):
    """
    Get Wallet Group Market Making Rewards History

    Limit: 10 requests per 1 second

    GET /v1/public/market_making_rewards/group_rewards_history

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-wallet-group-market-making-rewards-history
    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address, "symbol": symbol}
    return self._request("GET", "/v1/public/market_making_rewards/group_rewards_history", payload=payload)


def get_market_maker_current_epoch_estimate(self, address: str):
    """
    Get Market Maker Current Epoch Estimate

    Limit: 10 requests per 1 second

    GET /v1/public/market_making_rewards/current_epoch_estimate

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-market-maker-current-epoch-estimate
    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address}
    return self._request("GET", "/v1/public/market_making_rewards/current_epoch_estimate", payload=payload)


def get_wallet_staked_balance(self, address: str):
    """
    Get Wallet's Current Staked Balance

    Limit: 10 requests per 1 second

    GET /v1/staking/balance

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-wallets-current-staked-balance
    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address}
    return self._request("GET", "/v1/staking/balance", payload=payload)


def get_unstake_details(self, address: str):
    """
    Get Unstake Details

    Limit: 10 requests per 1 second

    GET /v1/staking/unstake_details

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-unstaking-order-details
    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address}
    return self._request("GET", "/v1/staking/unstake_details", payload=payload)



def get_staking_overview(self):
    """
    Get Wallet's Current Staked Balance

    Limit: 10 requests per 1 second

    GET /v1/staking/overview

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-staking-overview
    """
    return self._request("GET", "/v1/staking/overview")


def get_valor_batch_info(self):
    """
    Get Valor Batch Info

    Limit: 10 requests per 1 second

    GET /v1/staking/valor/batch_info

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-valor-batch-info
    """
    return self._request("GET", "/v1/staking/valor/batch_info")


def get_valor_pool_info(self):
    """
    Get Valor Pool Info

    Limit: 10 requests per 1 second

    GET /v1/staking/valor/pool_info

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-valor-pool-info
    """
    return self._request("GET", "/v1/staking/valor/pool_info")


def get_valor_redeem_info(self, address: str):
    """
    Get Valor Redeem Info

    Limit: 10 requests per 1 second

    GET /v1/staking/valor/redeem

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/get-valor-redeem-info
    """
    check_required_parameters([[address, "address"]])
    payload = {"address": address}
    return self._request("GET", "/v1/staking/valor/redeem", payload=payload)



from orderly_evm_connector.lib.utils import check_required_parameters
from orderly_evm_connector.lib.utils import get_withdraw_settle_verifyingcontract


def delegate_signer(
    self, delegateContract: str, brokerId: str, chainId: int, registrationNonce: int, txHash: str, timestamp: int, userAddress: str
):
    """
    Delegate Signer

    Limit: 1 requests per 1 second

    POST /v1/delegate_signer

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/add-delegate-signer
    """
    _message = {
        "delegateContract": delegateContract,
        "brokerId": brokerId,
        "chainId": chainId,
        "registrationNonce": registrationNonce,
        "txHash": txHash,
        "timestamp": timestamp
    }

    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
        },
        "message": _message,
        "primaryType": "DelegateSigner",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "DelegateSigner": [
                {"name": "delegateContract", "type": "address"},
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "timestamp", "type": "uint64"},
                {"name": "registrationNonce", "type": "uint256"},
                {"name": "txHash", "type": "bytes32"},
            ],
        },
    }

    _signature = self.get_wallet_signature(message=message)
    payload = {"message": _message, "signature": _signature, "userAddress": userAddress}
    check_required_parameters(
        [
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [registrationNonce, "registrationNonce"],
            [userAddress, "userAddress"],
        ]
    )
    return self._request("POST", "/v1/delegate_signer", payload=payload)


def delegate_add_orderly_key(
    self,
    delegateContract: str,
    brokerId: str,
    chainId: int,
    orderlyKey: str,
    scope: str,
    timestamp: int,
    expiration: int,
    userAddress: str,
    **kwargs
):
    """
    Delegate Orderly Key

    Limit: 1 requests per 1 second

    POST /v1/delegate_orderly_key

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/delegate_orderly_key
    """
    _message = {
        "delegateContract": delegateContract,
        "brokerId": brokerId,
        "chainId": chainId,
        "orderlyKey": orderlyKey,
        "scope": scope,
        "timestamp": timestamp,
        "expiration": expiration,
    }
    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
        },
        "message": _message,
        "primaryType": "DelegateAddOrderlyKey",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "AddOrderlyKey": [
                {"name": "delegateContract", "type": "address"},
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "orderlyKey", "type": "string"},
                {"name": "scope", "type": "string"},
                {"name": "timestamp", "type": "uint64"},
                {"name": "expiration", "type": "uint64"},
            ],
        },
    }
    _signature = self.get_wallet_signature(message=message)
    payload = {
        "message": _message,
        "signature": _signature,
        "userAddress": userAddress,
        **kwargs,
    }
    check_required_parameters(
        [
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [orderlyKey, "orderlyKey"],
            [scope, "scope"],
            [timestamp, "timestamp"],
            [expiration, "expiration"],
            [userAddress, "userAddress"],
        ]
    )
    return self._request("POST", "/v1/delegate_orderly_key", payload=payload)


def delegate_withdraw_request(
    self,
    delegateContract: int,
    userAddress: str,
    brokerId: str,
    chainId: int,
    receiver: str,
    token: str,
    amount: str,
    withdrawNonce: int,
    timestamp: int
):
    """
    Delegate Withdraw Request

    Limit: 1 requests per 1 second

    POST /v1/delegate_withdraw_request

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/delegate_withdraw_request
    """
    check_required_parameters(
        [
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [receiver, "receiver"],
            [token, "token"],
            [amount, "amount"],
            [withdrawNonce, "withdrawNonce"],
            [timestamp, "timestamp"],
            [delegateContract, "delegateContract"]
        ]
    )
    _message = {
        "brokerId": brokerId,
        "chainId": chainId,
        "receiver": receiver,
        "token": token,
        "amount": amount,
        "withdrawNonce": withdrawNonce,
        "timestamp": timestamp,
        "delegateContract": delegateContract,
    }
    verifyingContract = get_withdraw_settle_verifyingcontract(self.orderly_testnet)
    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": verifyingContract,
        },
        "message": _message,
        "primaryType": "DelegateWithdraw",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "DelegateWithdraw": [
                {"name": "delegateContract", "type": "string"},
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "receiver", "type": "address"},
                {"name": "token", "type": "string"},
                {"name": "amount", "type": "uint256"},
                {"name": "withdrawNonce", "type": "uint64"},
                {"name": "timestamp", "type": "uint64"},
            ],
        },
    }
    _signature = self.get_wallet_signature(message=message)
    payload = {
        "message": _message,
        "signature": _signature,
        "userAddress": userAddress,
        "verifyingContract": verifyingContract,
    }
    return self._sign_request("POST", "/v1/delegate_withdraw_request", payload=payload)


def delegate_request_pnl_settlement(
    self,
    delegateContract: int,
    brokerId: str,
    chainId: int,
    settleNonce: int,
    userAddress: str,
    timestamp: int
):
    """
    Delegate Orderly Key

    Limit: 1 requests per 1 second

    POST /v1/delegate_settle_pnl

    https://docs.orderly.network/build-on-evm/evm-api/restful-api/public/delegate_settle_pnl
    """
    check_required_parameters(
        [
            [delegateContract, "delegateContract"],
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [settleNonce, "settleNonce"],
            [userAddress, "userAdress"],
            [timestamp, "timestamp"],
        ]
    )
    verifyingContract = get_withdraw_settle_verifyingcontract(self.orderly_testnet)
    _message = {
        "delegateContract": delegateContract,
        "brokerId": brokerId,
        "chainId": chainId,
        "settleNonce": settleNonce,
        "timestamp": timestamp,
    }
    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": verifyingContract,
        },
        "message": _message,
        "primaryType": "DelegateSettlePnl",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "DelegateSettlePnl": [
                {"name": "delegateContract", "type": "address"},
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "settleNonce", "type": "uint64"},
                {"name": "timestamp", "type": "uint64"},
            ],
        },
    }

    _signature = self.get_wallet_signature(message=message)
    payload = {
        "message": _message,
        "signature": _signature,
        "userAddress": userAddress,
        "verifyingContract": verifyingContract,
    }
    return self._sign_request("POST", "/v1/delegate_settle_pnl", payload=payload)



def request_orderbook(self, type: str, symbol: str):
    """Request orderbook data
    https://docs-api-evm.orderly.network/?shell#websocket-api-public-request-orderbook
    """

    _message = {
        "id": self.wss_id,
        "event": "request",
        "params": {"type": type, "symbol": symbol},
    }
    self.send(_message)


def get_orderbook(self, topic: str):
    """{symbol}@orderbook depth 100 push every 1s

    https://docs-api-evm.orderly.network/?shell#websocket-api-public-orderbook
    """
    _message = {"id": self.wss_id, "event": "subscribe", "topic": topic}
    self.send_message_to_server(_message)


def get_orderbookupdate(self, topic: str):
    """{symbol}@orderbookupdate updated orderbook push every 200ms

    https://docs-api-evm.orderly.network/#websocket-api-public-order-book-update
    """
    _message = {"id": self.wss_id, "event": "subscribe", "topic": topic}
    self.send_message_to_server(_message)


def get_trade(self, topic: str):
    """Push interval: real-time push

    https://docs-api-evm.orderly.network/#websocket-api-public-trade
    """
    _message = {"id": self.wss_id, "event": "subscribe", "topic": topic}
    self.send_message_to_server(_message)


def get_24h_ticker(self, topic: str):
    """Push interval: 1s

    https://docs-api-evm.orderly.network/#websocket-api-public-24h-ticker
    """
    _message = {"id": self.wss_id, "event": "subscribe", "topic": topic}
    self.send_message_to_server(_message)


def get_24h_tickers(self):
    """Push interval: 1s

    https://docs-api-evm.orderly.network/#websocket-api-public-24h-tickers
    """
    _message = {"id": self.wss_id, "event": "subscribe", "topic": "tickers"}
    self.send_message_to_server(_message)


def get_bbo(self, topic: str):
    """Push interval: 10ms

    https://docs-api-evm.orderly.network/#websocket-api-public-bbo
    """
    _message = {"id": self.wss_id, "event": "subscribe", "topic": topic}
    self.send_message_to_server(_message)


def get_bbos(self):
    """Push interval: 1s
    https://docs-api-evm.orderly.network/#websocket-api-public-bbos
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": "bbos",
    }
    self.send_message_to_server(_message)


def get_kline(self, topic: str):
    """{time}: 1m/5m/15m/30m/1h/1d/1w/1M
    Push interval: 1s
    Name	Type	Required	Description
    id	    string	Y	        id generate by client
    event	string	Y	        subscribe/unsubscribe
    topic	string	N	        {symbol}@kline_{time}

    https://docs-api-evm.orderly.network/#websocket-api-public-k-line
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": topic,
    }
    self.send_message_to_server(_message)


def get_index_price(self, topic: str):
    """Push interval: 1s
    https://docs-api-evm.orderly.network/#websocket-api-public-index-price
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": topic,
    }
    self.send_message_to_server(_message)


def get_index_prices(self):
    """Push interval: 1s
    https://docs-api-evm.orderly.network/#websocket-api-public-index-prices
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": "indexprices",
    }
    self.send_message_to_server(_message)


def get_mark_price(self, topic: str):
    """Push interval: 1s
    https://docs-api-evm.orderly.network/#websocket-api-public-mark-price
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": topic,
    }
    self.send_message_to_server(_message)


def get_mark_prices(self):
    """Push interval: 1s
    https://docs-api-evm.orderly.network/#websocket-api-public-mark-prices
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": "markprices",
    }
    self.send_message_to_server(_message)


def get_open_interest(self, topic: str):
    """Push interval: push every 1 second if open interest change and 10 seconds force update even if no change.
    https://docs-api-evm.orderly.network/#websocket-api-public-open-interest
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": topic,
    }
    self.send_message_to_server(_message)


def get_estimated_funding_rate(self, topic: str):
    """Push interval: 15s.
    https://docs-api-evm.orderly.network/#websocket-api-public-estimated-funding-rate
    """
    _message = {
        "id": self.wss_id,
        "event": "subscribe",
        "topic": topic,
    }
    self.send_message_to_server(_message)


def get_liquidation_push(self):
    """Push interval: push on addition/removal/update from list within 1s.
    https://docs-api-evm.orderly.network/#websocket-api-public-liquidation-push
    """
    _message = {"id": self.wss_id, "event": "subscribe", "topic": "liquidation"}
    self.send_message_to_server(_message)



from orderly_evm_connector.api import API
from orderly_evm_connector.async_api import AsyncAPI


class Rest(API):
    def __init__(
        self,
        orderly_key=None,
        orderly_secret=None,
        wallet_secret=None,
        orderly_testnet=False,
        **kwargs
    ):
        self.orderly_testnet = orderly_testnet
        super().__init__(
            orderly_key, orderly_secret, wallet_secret, orderly_testnet, **kwargs
        )

    # account
    from orderly_evm_connector.rest._account import remove_orderly_key
    from orderly_evm_connector.rest._account import get_registration_nonce
    from orderly_evm_connector.rest._account import get_account_details
    from orderly_evm_connector.rest._account import get_account
    from orderly_evm_connector.rest._account import get_broker
    from orderly_evm_connector.rest._account import register_account
    from orderly_evm_connector.rest._account import get_orderly_key
    from orderly_evm_connector.rest._account import add_orderly_key
    from orderly_evm_connector.rest._account import update_leverage_configuration
    from orderly_evm_connector.rest._account import get_current_holdings
    from orderly_evm_connector.rest._account import get_account_information
    from orderly_evm_connector.rest._account import set_maintenance_config
    from orderly_evm_connector.rest._account import get_user_daily_statistics
    from orderly_evm_connector.rest._account import get_user_daily_volume
    from orderly_evm_connector.rest._account import get_user_volume_statistics
    from orderly_evm_connector.rest._account import get_current_orderlykey_info
    from orderly_evm_connector.rest._account import get_orderlykey_ip_restriction
    from orderly_evm_connector.rest._account import set_orderlykey_ip_restriction
    from orderly_evm_connector.rest._account import reset_orderlykey_ip_restriction
    from orderly_evm_connector.rest._account import get_position_history

    # broker
    from orderly_evm_connector.rest._broker import get_list_of_brokers
    from orderly_evm_connector.rest._broker import get_user_fee_tier
    from orderly_evm_connector.rest._broker import get_broker_daily_volume
    from orderly_evm_connector.rest._broker import get_default_broker_fee
    from orderly_evm_connector.rest._broker import update_user_fee_rate
    from orderly_evm_connector.rest._broker import reset_user_fee_rate
    from orderly_evm_connector.rest._broker import update_default_broker_fee
    from orderly_evm_connector.rest._broker import get_tvl_by_broker

    # general
    from orderly_evm_connector.rest._general import get_system_maintenance_status
    from orderly_evm_connector.rest._general import get_faucet_usdc
    from orderly_evm_connector.rest._general import get_exchange_info
    from orderly_evm_connector.rest._general import get_token_info
    from orderly_evm_connector.rest._general import get_available_symbols
    from orderly_evm_connector.rest._general import get_fee_futures_information
    from orderly_evm_connector.rest._general import get_leverage_configuration
    from orderly_evm_connector.rest._general import get_user_statistics
    from orderly_evm_connector.rest._general import get_market_volume_by_broker

    # liquidation
    from orderly_evm_connector.rest._liquidation import get_positions_under_liquidation
    from orderly_evm_connector.rest._liquidation import get_liquidated_positions_info
    from orderly_evm_connector.rest._liquidation import get_insurance_fund_info
    from orderly_evm_connector.rest._liquidation import (
        get_liquidated_positions_by_liquidator,
    )
    from orderly_evm_connector.rest._liquidation import (
        get_liquidated_positions_of_account,
    )
    from orderly_evm_connector.rest._liquidation import claim_liquidated_positions
    from orderly_evm_connector.rest._liquidation import claim_from_insurance_fund

    # market
    from orderly_evm_connector.rest._market import get_market_trades
    from orderly_evm_connector.rest._market import get_volume_statistics
    from orderly_evm_connector.rest._market import (
        get_predicted_funding_rate_for_all_markets,
    )
    from orderly_evm_connector.rest._market import (
        get_predicted_funding_rate_for_one_market,
    )
    from orderly_evm_connector.rest._market import (
        get_funding_rate_history_for_one_market,
    )
    from orderly_evm_connector.rest._market import get_futures_info_for_all_markets
    from orderly_evm_connector.rest._market import get_futures_info_for_one_market
    from orderly_evm_connector.rest._market import get_tradingview_configuration
    from orderly_evm_connector.rest._market import get_tradingview_history_basrs
    from orderly_evm_connector.rest._market import get_tradingview_symbol_info
    from orderly_evm_connector.rest._market import get_orderbook_snapshot
    from orderly_evm_connector.rest._market import get_kline

    # notifications
    from orderly_evm_connector.rest._notifications import get_all_notifications
    from orderly_evm_connector.rest._notifications import get_unread_notifications
    from orderly_evm_connector.rest._notifications import set_read_status_notifications
    from orderly_evm_connector.rest._notifications import (
        set_read_status_all_notifications,
    )

    # system
    from orderly_evm_connector.rest._system import get_valut_balances
    # from orderly_evm_connector.rest._system import get_valut_chain_config
    from orderly_evm_connector.rest._system import get_supported_chains_broker

    # settlement
    from orderly_evm_connector.rest._settlement import get_settle_pnl_nonce
    from orderly_evm_connector.rest._settlement import request_pnl_settlement
    from orderly_evm_connector.rest._settlement import get_pnl_settlement_history

    # trade
    from orderly_evm_connector.rest._trade import create_algo_order
    from orderly_evm_connector.rest._trade import create_order
    from orderly_evm_connector.rest._trade import batch_create_order
    from orderly_evm_connector.rest._trade import edit_algo_order
    from orderly_evm_connector.rest._trade import edit_order
    from orderly_evm_connector.rest._trade import cancel_algo_order
    from orderly_evm_connector.rest._trade import cancel_algo_all_pending_order
    from orderly_evm_connector.rest._trade import cancel_order
    from orderly_evm_connector.rest._trade import cancel_algo_order_by_client_order_id
    from orderly_evm_connector.rest._trade import cancel_order_by_client_order_id
    from orderly_evm_connector.rest._trade import cancel_orders
    from orderly_evm_connector.rest._trade import batch_cancel_orders
    from orderly_evm_connector.rest._trade import batch_cancel_orders_by_client_order_id
    from orderly_evm_connector.rest._trade import get_algo_order
    from orderly_evm_connector.rest._trade import get_algo_orders
    from orderly_evm_connector.rest._trade import get_order
    from orderly_evm_connector.rest._trade import get_algo_order_by_client_order_id
    from orderly_evm_connector.rest._trade import get_order_by_client_order_id
    from orderly_evm_connector.rest._trade import get_orders
    from orderly_evm_connector.rest._trade import get_all_trades_of_order
    from orderly_evm_connector.rest._trade import get_trades
    from orderly_evm_connector.rest._trade import get_trade
    from orderly_evm_connector.rest._trade import get_all_positions_info
    from orderly_evm_connector.rest._trade import get_one_position_info
    from orderly_evm_connector.rest._trade import get_funding_fee_history

    # wallet
    from orderly_evm_connector.rest._wallet import get_asset_history
    from orderly_evm_connector.rest._wallet import get_withdraw_nonce
    from orderly_evm_connector.rest._wallet import withdraw_request
    from orderly_evm_connector.rest._wallet import internal_transfer

    #campaign
    from orderly_evm_connector.rest._campaign import get_points_epoch
    from orderly_evm_connector.rest._campaign import get_points_epochdates
    from orderly_evm_connector.rest._campaign import get_user_points
    from orderly_evm_connector.rest._campaign import get_points_leaderboard
    from orderly_evm_connector.rest._campaign import get_tradingrewards_epoch
    from orderly_evm_connector.rest._campaign import get_campaign_user_info
    #referral
    from orderly_evm_connector.rest._referral import bind_referral_code
    from orderly_evm_connector.rest._referral import get_referral_history
    from orderly_evm_connector.rest._referral import get_referral_code_info
    from orderly_evm_connector.rest._referral import get_distribution_history
    from orderly_evm_connector.rest._referral import check_ref_code
    from orderly_evm_connector.rest._referral import verify_ref_code
    from orderly_evm_connector.rest._referral import get_referral_rebate_summary
    from orderly_evm_connector.rest._referral import get_referral_info
    from orderly_evm_connector.rest._referral import get_referee_info
    from orderly_evm_connector.rest._referral import get_referee_history
    from orderly_evm_connector.rest._referral import edit_referral_split
    from orderly_evm_connector.rest._referral import create_referral_code
    from orderly_evm_connector.rest._referral import update_referral_code
    # rewards
    from orderly_evm_connector.rest._rewards import get_parameters_of_each_epoch
    from orderly_evm_connector.rest._rewards import get_broker_allocation_history
    from orderly_evm_connector.rest._rewards import get_account_trading_rewards_history
    from orderly_evm_connector.rest._rewards import get_wallet_trading_rewards_history
    from orderly_evm_connector.rest._rewards import get_epochs_data
    from orderly_evm_connector.rest._rewards import get_current_epoch_estimate
    from orderly_evm_connector.rest._rewards import get_current_epoch_estimate_broker
    from orderly_evm_connector.rest._rewards import get_parameters_of_each_mm_epoch
    from orderly_evm_connector.rest._rewards import get_wallet_group_mm_rewards_history
    from orderly_evm_connector.rest._rewards import get_market_maker_current_epoch_estimate
    from orderly_evm_connector.rest._rewards import get_wallet_staked_balance
    from orderly_evm_connector.rest._rewards import get_staking_overview
    from orderly_evm_connector.rest._rewards import get_unstake_details
    from orderly_evm_connector.rest._rewards import get_valor_batch_info
    from orderly_evm_connector.rest._rewards import get_valor_pool_info
    from orderly_evm_connector.rest._rewards import get_valor_redeem_info


class RestAsync(AsyncAPI):
    def __init__(
        self,
        orderly_key=None,
        orderly_secret=None,
        wallet_secret=None,
        orderly_testnet=False,
        **kwargs
    ):
        self.orderly_testnet = orderly_testnet
        super().__init__(
            orderly_key, orderly_secret, wallet_secret, orderly_testnet, **kwargs
        )

    # account
    from orderly_evm_connector.rest._account import get_registration_nonce
    from orderly_evm_connector.rest._account import get_account_details
    from orderly_evm_connector.rest._account import get_account
    from orderly_evm_connector.rest._account import get_broker
    from orderly_evm_connector.rest._account import register_account
    from orderly_evm_connector.rest._account import get_orderly_key
    from orderly_evm_connector.rest._account import add_orderly_key
    from orderly_evm_connector.rest._account import update_leverage_configuration
    from orderly_evm_connector.rest._account import get_current_holdings
    from orderly_evm_connector.rest._account import get_account_information
    from orderly_evm_connector.rest._account import set_maintenance_config
    from orderly_evm_connector.rest._account import get_user_daily_statistics
    from orderly_evm_connector.rest._account import get_user_daily_volume
    from orderly_evm_connector.rest._account import get_user_volume_statistics
    from orderly_evm_connector.rest._account import get_current_orderlykey_info
    from orderly_evm_connector.rest._account import get_orderlykey_ip_restriction
    from orderly_evm_connector.rest._account import set_orderlykey_ip_restriction
    from orderly_evm_connector.rest._account import reset_orderlykey_ip_restriction
    from orderly_evm_connector.rest._account import get_position_history

    from orderly_evm_connector.rest._broker import get_default_broker_fee
    from orderly_evm_connector.rest._broker import update_user_fee_rate
    from orderly_evm_connector.rest._broker import reset_user_fee_rate
    from orderly_evm_connector.rest._broker import update_default_broker_fee
    from orderly_evm_connector.rest._account import remove_orderly_key

    # broker
    from orderly_evm_connector.rest._broker import get_list_of_brokers
    from orderly_evm_connector.rest._broker import get_user_fee_tier
    from orderly_evm_connector.rest._broker import get_broker_daily_volume

    # general
    from orderly_evm_connector.rest._general import get_system_maintenance_status
    from orderly_evm_connector.rest._general import get_faucet_usdc
    from orderly_evm_connector.rest._general import get_exchange_info
    from orderly_evm_connector.rest._general import get_token_info
    from orderly_evm_connector.rest._general import get_available_symbols
    from orderly_evm_connector.rest._general import get_fee_futures_information
    from orderly_evm_connector.rest._general import get_leverage_configuration
    from orderly_evm_connector.rest._general import get_user_statistics

    # liquidation
    from orderly_evm_connector.rest._liquidation import get_positions_under_liquidation
    from orderly_evm_connector.rest._liquidation import get_liquidated_positions_info
    from orderly_evm_connector.rest._liquidation import get_insurance_fund_info
    from orderly_evm_connector.rest._liquidation import (
        get_liquidated_positions_by_liquidator,
    )
    from orderly_evm_connector.rest._liquidation import (
        get_liquidated_positions_of_account,
    )
    from orderly_evm_connector.rest._liquidation import claim_liquidated_positions
    from orderly_evm_connector.rest._liquidation import claim_from_insurance_fund

    # market
    from orderly_evm_connector.rest._market import get_market_trades
    from orderly_evm_connector.rest._market import get_volume_statistics
    from orderly_evm_connector.rest._market import (
        get_predicted_funding_rate_for_all_markets,
    )
    from orderly_evm_connector.rest._market import (
        get_predicted_funding_rate_for_one_market,
    )
    from orderly_evm_connector.rest._market import (
        get_funding_rate_history_for_one_market,
    )
    from orderly_evm_connector.rest._market import get_futures_info_for_all_markets
    from orderly_evm_connector.rest._market import get_futures_info_for_one_market
    from orderly_evm_connector.rest._market import get_tradingview_configuration
    from orderly_evm_connector.rest._market import get_tradingview_history_basrs
    from orderly_evm_connector.rest._market import get_tradingview_symbol_info
    from orderly_evm_connector.rest._market import get_orderbook_snapshot
    from orderly_evm_connector.rest._market import get_kline

    # notifications
    from orderly_evm_connector.rest._notifications import get_all_notifications
    from orderly_evm_connector.rest._notifications import get_unread_notifications
    from orderly_evm_connector.rest._notifications import set_read_status_notifications
    from orderly_evm_connector.rest._notifications import (
        set_read_status_all_notifications,
    )

    # system
    from orderly_evm_connector.rest._system import get_valut_balances
    from orderly_evm_connector.rest._system import get_supported_chains_broker

    # settlement
    from orderly_evm_connector.rest._settlement import get_settle_pnl_nonce
    from orderly_evm_connector.rest._settlement import request_pnl_settlement
    from orderly_evm_connector.rest._settlement import get_pnl_settlement_history

    # trade
    from orderly_evm_connector.rest._trade import create_algo_order
    from orderly_evm_connector.rest._trade import create_order
    from orderly_evm_connector.rest._trade import batch_create_order
    from orderly_evm_connector.rest._trade import edit_algo_order
    from orderly_evm_connector.rest._trade import edit_order
    from orderly_evm_connector.rest._trade import cancel_algo_order
    from orderly_evm_connector.rest._trade import cancel_algo_all_pending_order
    from orderly_evm_connector.rest._trade import cancel_order
    from orderly_evm_connector.rest._trade import cancel_algo_order_by_client_order_id
    from orderly_evm_connector.rest._trade import cancel_order_by_client_order_id
    from orderly_evm_connector.rest._trade import cancel_orders
    from orderly_evm_connector.rest._trade import batch_cancel_orders
    from orderly_evm_connector.rest._trade import batch_cancel_orders_by_client_order_id
    from orderly_evm_connector.rest._trade import get_algo_order
    from orderly_evm_connector.rest._trade import get_algo_orders
    from orderly_evm_connector.rest._trade import get_order
    from orderly_evm_connector.rest._trade import get_algo_order_by_client_order_id
    from orderly_evm_connector.rest._trade import get_order_by_client_order_id
    from orderly_evm_connector.rest._trade import get_orders
    from orderly_evm_connector.rest._trade import get_all_trades_of_order
    from orderly_evm_connector.rest._trade import get_trades
    from orderly_evm_connector.rest._trade import get_trade
    from orderly_evm_connector.rest._trade import get_all_positions_info
    from orderly_evm_connector.rest._trade import get_one_position_info
    from orderly_evm_connector.rest._trade import get_funding_fee_history

    # wallet
    from orderly_evm_connector.rest._wallet import get_asset_history
    from orderly_evm_connector.rest._wallet import get_withdraw_nonce
    from orderly_evm_connector.rest._wallet import withdraw_request
    from orderly_evm_connector.rest._wallet import internal_transfer

    # campaign
    from orderly_evm_connector.rest._campaign import get_points_epoch
    from orderly_evm_connector.rest._campaign import get_points_epochdates
    from orderly_evm_connector.rest._campaign import get_user_points
    from orderly_evm_connector.rest._campaign import get_points_leaderboard
    from orderly_evm_connector.rest._campaign import get_tradingrewards_epoch
    from orderly_evm_connector.rest._campaign import get_campaign_user_info
    # referral
    from orderly_evm_connector.rest._referral import bind_referral_code
    from orderly_evm_connector.rest._referral import get_referral_history
    from orderly_evm_connector.rest._referral import get_referral_code_info
    from orderly_evm_connector.rest._referral import get_distribution_history
    from orderly_evm_connector.rest._referral import check_ref_code
    from orderly_evm_connector.rest._referral import verify_ref_code
    from orderly_evm_connector.rest._referral import get_referral_rebate_summary
    from orderly_evm_connector.rest._referral import get_referral_info
    from orderly_evm_connector.rest._referral import get_referee_info
    from orderly_evm_connector.rest._referral import get_referee_history
    from orderly_evm_connector.rest._referral import edit_referral_split
    from orderly_evm_connector.rest._referral import create_referral_code
    from orderly_evm_connector.rest._referral import update_referral_code

    # rewards
    from orderly_evm_connector.rest._rewards import get_parameters_of_each_epoch
    from orderly_evm_connector.rest._rewards import get_broker_allocation_history
    from orderly_evm_connector.rest._rewards import get_account_trading_rewards_history
    from orderly_evm_connector.rest._rewards import get_wallet_trading_rewards_history
    from orderly_evm_connector.rest._rewards import get_epochs_data
    from orderly_evm_connector.rest._rewards import get_current_epoch_estimate
    from orderly_evm_connector.rest._rewards import get_current_epoch_estimate_broker
    from orderly_evm_connector.rest._rewards import get_parameters_of_each_mm_epoch
    from orderly_evm_connector.rest._rewards import get_wallet_group_mm_rewards_history
    from orderly_evm_connector.rest._rewards import get_market_maker_current_epoch_estimate
    from orderly_evm_connector.rest._rewards import get_wallet_staked_balance
    from orderly_evm_connector.rest._rewards import get_staking_overview
    from orderly_evm_connector.rest._rewards import get_unstake_details
    from orderly_evm_connector.rest._rewards import get_valor_batch_info
    from orderly_evm_connector.rest._rewards import get_valor_pool_info
    from orderly_evm_connector.rest._rewards import get_valor_redeem_info



from orderly_evm_connector.lib.utils import check_required_parameters


def get_list_of_brokers(self, broker_id: str = None):
    """Get list of brokers

    Limit: 10 requests per 1 second per IP address

    GET /v1/public/broker/name?broker_id=:broker_id

    Get the list of brokers currently onboarded on Orderly Network.

    Args:
        broker_id(string): If provided, it will only output details for the particular broker.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/public/get-broker-list
    """
    payload = {"broker_id": broker_id}
    return self._request("GET", "/v1/public/broker/name", payload=payload)

#add Get User Fee Tier API in Broker
def get_user_fee_tier(self,account_id: str = None,address: str = None,page: int = None,size: int = None):
    """Get the user fee rate information. Only address or account_id should be provided, not both.

    Limit 10 requests per 60 seconds

    GET /v1/broker/user_info
    Optional Aargs:
        account_id(string)
        
        address(string)
        
        page(number)
        
        size(number)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-user-fee-tier
    """
    payload = {"account_id": account_id,"address":address,"page":page,"size":size}
    return self._sign_request("GET", "/v1/broker/user_info", payload=payload)

def get_broker_daily_volume(
    self, 
    start_date: str,
    end_date: str,
    broker_id: str = None,
    address: str = None,
    order_tags: str = None,
    aggregateBy: str = None,
    sort: str = None
    ):
    """Get Broker Daily Volume
    Limit 10 requests per 60 seconds

    GET /v1/volume/broker/daily

    Get the daily historical breakdown of the user trading volume on specified broker.
    The provided start_date/end_date has to be within a 90-day range.

    Args:
        start_date(string): Format YYYY-MM-DD.
        end_date(string): Format YYYY-MM-DD.
    Optional Args:
        broker_id(string)
        address(string)
        order_tags(string)
        aggregateBy(string)
        sort(string)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-broker-daily-volume
    """
    check_required_parameters([[start_date, "start_date"], [end_date, "end_date"]])
    payload = {
        "start_date": start_date,
        "end_date": end_date,
        "broker_id": broker_id,
        "address": address,
        "order_tags": order_tags,
        "aggregateBy": aggregateBy,
        "sort": sort
        }
    return self._sign_request("GET", "/v1/volume/broker/daily", payload=payload)


def get_user_fee_rates(self, account_id: str = None, address: str = None, page: int = None, size: int = None):
    payload = {
        "account_id": account_id,
        "address": address,
        "page": page,
        "size": size
    }
    return self._sign_request("GET", "/v1/broker/user_info", payload=payload)


def update_user_fee_rate(self, account_ids: str, maker_fee_rate: float, taker_fee_rate: float):
    check_required_parameters(
        [[account_ids, "account_ids"],
         [maker_fee_rate, "maker_fee_rate"],
         [taker_fee_rate, "taker_fee_rate"]]
    )
    payload = {
        "account_ids": account_ids,
        "maker_fee_rate": maker_fee_rate,
        "taker_fee_rate": taker_fee_rate
    }
    return self._sign_request("POST", "/v1/broker/fee_rate/set", payload=payload)


def reset_user_fee_rate(self, account_ids: str):
    check_required_parameters(
        [[account_ids, "account_ids"]]
    )
    payload = {
        "account_ids": account_ids
    }
    return self._sign_request("POST", "/v1/broker/fee_rate/set_default", payload=payload)


def update_default_broker_fee(self, maker_fee_rate: float, taker_fee_rate: float):
    check_required_parameters(
        [[maker_fee_rate, "maker_fee_rate"],
         [taker_fee_rate, "taker_fee_rate"]]
    )
    payload = {
        "maker_fee_rate": maker_fee_rate,
        "taker_fee_rate": taker_fee_rate
    }
    return self._sign_request("POST", "/v1/broker/fee_rate/default", payload=payload)


def get_default_broker_fee(self):
    return self._sign_request("GET", "/v1/broker/fee_rate/default")


def get_tvl_by_broker(self, broker_id: str = None):
    payload = {
        "broker_id": broker_id
    }
    return self._sign_request("GET", "/v1/public/balance/stats", payload=payload)


from orderly_evm_connector.lib.utils import check_required_parameters, get_timestamp
from orderly_evm_connector.lib.utils import check_enum_parameter,get_withdraw_settle_verifyingcontract
from orderly_evm_connector.lib.enums import WalletSide, AssetStatus


def get_asset_history(
    self,
    token: str = None,
    side: str = None,
    status: str = None,
    start_t: int = None,
    end_t: int = None,
    page: int = None,
    size: int = None,
):
    """Get asset history

    Limit 10 requests per 60 seconds

    GET /v1/asset/history

    Get asset history, including token deposits/withdrawals.

    Optional Args:
        token(string):	            token name you want to search
        side(string):	            DEPOSIT/WITHDRAW
        status(string):	        NEW/CONFIRM/PROCESSING/COMPLETED/FAILED
        start_t(timestamp):	    start time range that wish to query, noted the time stamp is 13-digits timestamp.
        end_t(timestamp):       end time range that wish to query, noted the time stamp is 13-digits timestamp.
        page(number):           (default: 1)	the page you wish to query.
        size(number):           (default: 25)

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-asset-history
    """
    if status:
        check_enum_parameter(status, AssetStatus)
    if side:
        check_enum_parameter(side, WalletSide)

    payload = {
        "token": token,
        "side": side,
        "status": status,
        "start_t": start_t,
        "end_t": end_t,
        "page": page,
        "size": size,
    }

    return self._sign_request("GET", "/v1/asset/history", payload=payload)


def get_withdraw_nonce(self):
    """Get Withdrawal Nonce
    Limit 10 requests per 1 seconds

    GET /v1/withdraw_nonce

    Retrieve a nonce used for requesting a withdrawal on Orderly Network. Each nonce can only be used once.

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/get-withdrawal-nonce
    """

    return self._sign_request("GET", "/v1/withdraw_nonce")


def withdraw_request(
    self,
    brokerId: str,
    chainId: int,
    receiver: str,
    token: str,
    amount: str,
    withdrawNonce: int,
    userAddress: str,
    # verifyingContract: str,
):
    """Create Withdraw Request
    
    Limit: 10 requests per 1 second per IP address

    POST /v1/withdraw_request
    
    This API will throw an error with message 22 - Cross-chain withdrawal required for this withdrawal request. if allow_cross_chain_withdrawal is false while the request is a cross-chain request.

    Args
    message(JSON): Message object containing the message that is signed by the wallet owner
    message.brokerId(string): Broker ID
    message.allowCrossChainWithdrawal(boolean)
    message.chainId(int): Chain ID of registering chain (within those that are supported by the Network)
    message.receiver(string): Address of the receiver, which should be equal to the address of the account.
    message.token(string): The string representation of the token that is being withdrawn (eg "USDC")
    message.amount(int): Amount of tokens to be withdrawn
    message.withdrawNonce(int): Valid withdrawal nonce from Get withdrawal nonce
    message.timestamp(timestamp): current timestamp in UNIX milliseconds
    signature(string): The signature generated by signing the message object via EIP-712
    userAddress(string): The address of the wallet signing the message object via EIP-712
    # verifyingContract(string): Address of the Orderly Network ledger contract

    https://orderly.network/docs/build-on-evm/evm-api/restful-api/private/create-withdraw-request

    """
    check_required_parameters(
        [
            [brokerId, "brokerId"],
            [chainId, "chainId"],
            [receiver, "receiver"],
            [token, "token"],
            [amount, "amount"],
            [withdrawNonce, "withdrawNonce"],
            [userAddress, "userAddress"],
            # [verifyingContract, "verifyingContract"],
        ]
    )
    _message = {
        "brokerId": brokerId,
        "chainId": chainId,
        "receiver": receiver,
        "token": token,
        "amount": amount,
        "withdrawNonce": withdrawNonce,
        "timestamp": int(get_timestamp()),
        "type": "Withdraw",
    }
    verifyingContract = get_withdraw_settle_verifyingcontract(self.orderly_testnet)
    message = {
        "domain": {
            "name": "Orderly",
            "version": "1",
            "chainId": chainId,
            "verifyingContract": verifyingContract,
        },
        "message": _message,
        "primaryType": "Withdraw",
        "types": {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            "Withdraw": [
                {"name": "brokerId", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "receiver", "type": "address"},
                {"name": "token", "type": "string"},
                {"name": "amount", "type": "uint256"},
                {"name": "withdrawNonce", "type": "uint64"},
                {"name": "timestamp", "type": "uint64"},
            ],
        },
    }
    _signature = self.get_wallet_signature(message=message)
    payload = {
        "message": _message,
        "signature": _signature,
        "userAddress": userAddress,
        "verifyingContract": verifyingContract,
    }
    return self._sign_request("POST", "/v1/withdraw_request", payload=payload)


def internal_transfer(
    self,
    token: str,
    receiver_list: list,
):
    """Create Internal Transfer

    POST /v1/internal_transfer

    This API creates an internal transfer between accounts.

    Args:
    token(string): The string representation of the token to be transferred (e.g., "USDC")
    receiver_list(list): A list of dictionaries, each containing:
        - account_id (str): The account ID of the receiver
        - amount (float): The amount to be transferred to the receiver

    https://orderly.network/docs/build-on-omnichain/evm-api/restful-api/private/create-internal-transfer
    """
    check_required_parameters(
        [
            [token, "token"],
            [receiver_list, "receiver_list"],
        ]
    )
    
    for receiver in receiver_list:
        check_required_parameters(
            [
                [receiver.get("account_id"), "account_id"],
                [receiver.get("amount"), "amount"],
            ]
        )

    payload = {
        "token": token,
        "receiver_list": receiver_list,
    }

    return self._sign_request("POST", "/v1/internal_transfer", payload=payload)
